(ns prism)
(use 'clojure.contrib.duck-streams)

;; === What serves has a line accumulator
(def accu (atom ""))
(def current-state (atom [:root]))
(def current-style (atom ["Root"]))

; == Useful stuff I've taken from Stackoverflow
;; https://stackoverflow.com/questions/2897587/escape-sanitize-user-input-in-clojure-compojure
(defn- xml-str
   "Like clojure.core/str but escapes < > and &."
   [x]
   (-> x str (.replace "&" "&amp;") (.replace "<" "&lt;") (.replace ">" "&gt;")))

(defn read-file [f] (read-lines f))

;; Load the syntax rules of the language to highlight (required hash-map)
;; Maybe there's a better way to handle this ?
(def rulez (load-file "langs/lang_clojure.clj"))

(defn find-groups 
  "Given a regexp that matched against a string, returns a vector
  of the different matching parts."
  [match]
  (let [res (atom [])
        gc (. match groupCount)]
  (if (= 0 gc)
    (reset! res [(. match group 0)])                    ;; then
    (dotimes [i gc]                                     ;; else
      (reset! res (conj @res (. match group (+ 1 i))))
  ))
  @res
  ))

(defn handle-actions [bestmatch line]
  (let [act (:action bestmatch)]
    (if-not (= "pop" act)
      (do
        ;; (println (format "\n'%s'\nwith current-style %s\n Action %s\n" line (last @current-style) act))
        (if-not (= (last @current-style) act)
          (do
            (reset! current-style (conj @current-style act))
          )
          )
        ))))

(defn handle-pops [m line]
  (if (= (:action m) "pop")
    (do
      (if (> (.length @current-state) 1) (reset! current-state (pop @current-state)))
      (if (> (.length @current-style) 1) (reset! current-style (pop @current-style)))
    ))
  (if (:to-state m) ;; there is a state change here (not (= nil cb))
    (reset! current-state (conj @current-state (:to-state m)))
  )
  (if (:action m) (handle-actions m line)) ;; there is a style change here
)

(defn write-tag [x y]
  ;; (println (format "X: %s -- Y: %s" x y))
  (cond (= x "Root") (xml-str y)
        (= x "Url") (format "<span class=\"%s\"><a href=\"%s\">%s</a></span>" x y (xml-str y))
        :else (format "<span class=\"%s\">%s</span>" x (xml-str y))
  )
)

(defn format-tag [m style]
  (if (= "" m) 
    ""
    (write-tag style (xml-str m))
  ))
    
(defn format-tags [bef styles matches]
  (if-not (= "" bef)
    ;; there is something before the match
    ;; (apply str (write-tag (last @current-style) bef) (map write-tag styles matches))
    (apply str (write-tag (last @current-style) bef) (map write-tag styles matches))
    ;; there is nothing before the match
    (apply str (map write-tag styles matches))
  ))

(defn find-best-match
  "Given a hash-map whose keys are regexp and callbacks as values,
  find the 'best' match in a given string, ie the one that matched with the
  lowest index, or nothing."
  [rules string]
  (def bestmatch (atom {:re nil 
                        :start (. string length) 
                        :end 0 
                        :to-state nil 
                        :apply-style [@current-style] 
                        :action nil 
                        :groups [string]}))
  (doseq [v rules]
    (let [m (.matcher (:re v) string)]
      (if (. m find) (do
        (if (or (< (. m start 0) (:start @bestmatch)) 
                (and (= (. m start 0) (:start @bestmatch))
                 (> (. m end 0) (:end @bestmatch)))
            )
          (swap! bestmatch assoc 
                          :start (. m start 0) 
                          :end (. m end 0) 
                          :re (:re v) 
                          :to-state (:to-state v ) 
                          :apply-style (:apply-style v) 
                          :action (:action v) 
                          :groups (find-groups m))
          )))))
  @bestmatch
  )

(defn get-line-result
  "This routine returns a hash-map (Called by process-line)"
  [line]
  (let [bm (find-best-match ((last @current-state) rulez) line)
        re (bm :re)
        ms (bm :start)
        me (bm :end)
        ft (bm :apply-style)
        ts (bm :to-state)  ;; state change
        gp (bm :groups)    ;; captured groups
        act (bm :action)   ;; pop, etc.
        ]
    (if re
      ;; then we matched something
      (do
        (comment 
        (if ts ;; there is a state change here (not (= nil cb))
            (reset! current-state (conj @current-state ts))
        )
        (if act (handle-actions bm line)) ;; there is a style change here
      )
        {:before (. line substring 0 ms)
         :matched (. line substring ms me)
         :after (. line substring me (. line length))
         :apply-style ft
         :to-state ts
         :action act
         :groups gp
        }
      )
      ;; We didn't match, so returns the entire line
      {:before line
       :matched nil
       :after nil
       :apply-style ft ;  (last @current-style) ;
       :to-state ts
       :action act
       :groups gp
      }
    )))

(defn process-line
  "Reccursive procedure responsible of highlighting a given line"
  [text]
  (let [m (get-line-result text)]   
    (if (m :matched)
      (do 
        (reset! accu (str @accu (format-tags (:before m) (:apply-style m) (:groups m))))
        (handle-pops m text)
        (recur (m :after))
      )
      (do
        (handle-pops m text)
        (str @accu (format-tag (m :before) (last @current-style)))
      )
    )))

(defn highlight [text]
  (def res (atom [])) 
  (doseq [line (.split text "\n")]
    (do
        (reset! res (conj @res (process-line line)))
        ;; (println (format "'%s' --> %s STATES: %s" line @current-style @current-state))
        (reset! accu "") 
    ) 
  ) 
  (apply str (interpose "\n" @res))
)

(defn write-from-template [input-filename]
  (let [
    f (slurp "template.txt")
    extens ".clj"
    to-parse (highlight  (slurp (str input-filename extens)))]
    (clojure.contrib.duck-streams/spit (str input-filename ".html") (format f to-parse)))
  (println (format "Your %s.html file has been generated" input-filename))
) 

(write-from-template "prism")